import 'package:lambda_calculus/lambda_calculus.dart';
import 'package:test/test.dart';

void main() {
  test('Playground', () {
    final l1 = r'(\x \y \z. x z (y z)) (\a\b a)'.toLambda()!;
    print(l1.eval(evalType: LambdaEvaluationType.fullReduction));
  });

  test(
    'Clone Test',
    () {
      expect(
        identityHashCode(LambdaConstants.lambdaTrue),
        isNot(identityHashCode(LambdaConstants.lambdaTrue.clone())),
      );
      expect(LambdaConstants.lambdaTrue, LambdaConstants.lambdaTrue.clone());
      expect(
        identityHashCode(LambdaConstants.lambdaSeven),
        isNot(identityHashCode(LambdaConstants.lambdaSeven.clone())),
      );
      expect(LambdaConstants.lambdaSeven, LambdaConstants.lambdaSeven.clone());
      expect(
        identityHashCode(LambdaConstants.yCombinator),
        isNot(identityHashCode(LambdaConstants.yCombinator.clone())),
      );
      expect(LambdaConstants.yCombinator, LambdaConstants.yCombinator.clone());
    },
  );

  test('Parse Test', () {
    expect(
      r'\x1(λx2. x1 (λx3. x2 x2 x3)) (λx2. x1 (λ x2 x2 0))'.toLambda(),
      LambdaConstants.yCombinator,
    );
    expect(
      r'\_x1 _x1'.toLambda(),
      LambdaConstants.lambdaIdentity,
    );
    expect(r'\_x2 _x1'.toLambda(), null);
    expect(r'((x1 (\x2 x2))'.toLambda(), null);
  });

  test(
    'Evaluation Test',
    () {
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaTest,
          LambdaConstants.lambdaFalse,
          LambdaConstants.lambdaTwo,
          LambdaConstants.lambdaOne,
        ]).eval(),
        LambdaConstants.lambdaOne,
      );
      expect(LambdaConstants.yCombinator.eval(), LambdaConstants.yCombinator);
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaSucc,
          LambdaConstants.lambdaSix,
        ]).toInt(),
        7,
      );
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaPlus,
          LambdaConstants.lambdaTwo,
          LambdaConstants.lambdaThree,
        ]).toInt(),
        5,
      );
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaTimes,
          LambdaConstants.lambdaTwo,
          LambdaConstants.lambdaThree,
        ]).toInt(),
        6,
      );
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaPower,
          LambdaConstants.lambdaTwo,
          LambdaConstants.lambdaThree,
        ]).toInt(),
        8,
      );
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaIsZero,
          LambdaConstants.lambdaTwo,
        ]).eval(),
        LambdaConstants.lambdaFalse,
      );
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaIsZero,
          LambdaConstants.lambdaZero,
        ]).eval(),
        LambdaConstants.lambdaTrue,
      );
      expect(
        Lambda.applyAll([
          LambdaConstants.lambdaAnd,
          LambdaConstants.lambdaFalse,
          LambdaConstants.omega,
        ]).eval(evalType: LambdaEvaluationType.callByName),
        LambdaConstants.lambdaFalse,
      );
    },
  );
}
